# Rain Droplet Community Credit Software
# Copyright (C) 2006-11
# 
# This program is free software; you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the 
# Free Software Foundation; either version 2 of the License, or (at your 
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program; if not, write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Based on RippleSite by Ryan Fugger
# ryan@ripplepay.com
#-------------------------------------------------------------------------

"""
Django views for connection accounts
"""

from django.core import template_loader
from django.core.extensions import DjangoContext
from django.utils.httpwrappers import HttpResponse, HttpResponseRedirect
import django.conf.settings as settings
from django.core.meta import Q
from django.core.db import db
from django.utils.html import escape

from django.models.droplet import *
from raindroplet.droplet.views import checkLogin, getAccountsSummary, parseEmails, getEmails, Struct, sendEmail, getSessionInfos, getRateProposalAccts
import routing
from pathgraph import computeInterest

import datetime


def accountList(request):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  
  d = {}
  d['infos'] = getSessionInfos(request)
  
  d['accountCurrencies'] = getAccountsSummary(userNode)
  
   # outstanding offers of credit received/made
  d['receivedOffers'] = offers.get_list(recipient_email__in=getEmails(userNode), order_by=('-id',))
  d['sentOffers'] = userNode.get_sent_offer_list(order_by=('-id',))
  d['rateProposals'] = getRateProposalAccts(userNode)
  t = template_loader.get_template('accountList')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

def accountDetail(request, acctId):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)

  if accounts.get_count(pk=acctId) > 0:
    acct = accounts.get_object(pk=acctId)
  else: # can't access accounts that aren't in database and active
    return HttpResponseRedirect('/accounts/')
  
  if acct.owner_id != userNode.id: # can't access accounts that aren't yours!
    return HttpResponseRedirect('/accounts/')
  
  d = {}
  d['infos'] = getSessionInfos(request)
  
  if request.POST: # modify account or set error
    if request.POST['name'] == '':
      d['infos'].append("Error: Account name can't be left blank.")
    elif acct.name != request.POST['name']:
      acct.name = request.POST['name']
      d['infos'].append("Account name changed.")
    
    try:
      interest_rate = float(request.POST['interest']) / 100.0 # may raise ValueError
      shared_data = acct.get_shared_data()
      if interest_rate <> shared_data.interest_rate:
        if shared_data.proposed_rate and interest_rate == shared_data.proposed_rate and userNode.id == shared_data.node_to_confirm_id:
          # take manually setting to otherly-proposed rate to be acceptance
          accept_proposed_rate(request, shared_data)
          d['infos'].append("Interest rate set to %s%%." % (interest_rate * 100.0))
        else:
          info = propose_rate(request, shared_data, interest_rate, node_to_confirm=acct.get_partner())
          d['infos'].append(info)
          
    except ValueError:
      d['infos'].append("Interest rate must be a number.")
    
    try: # modify partner's max credit limit
      partner_limit = float(request.POST['partner_limit'])
      if partner_limit < 0.0:
        d['infos'].append("Error: Partner's credit limit must be a positive number or zero. '%s' is invalid." % request.POST['partner_limit'])
        raise ValueError
      elif partner_limit >= 1e12:
        d['infos'].append("Error: Partner limit amount is too large to be stored in the database.  Please enter another amount.")
        raise ValueError
      elif partner_limit <= 1e-12 and partner_limit != 0.0:
        d['infos'].append("Error: Partner limit amount is too small to be stored in the database.  Please enter another amount.")
        raise ValueError
      
      if acct.partner_limit != partner_limit:
        acct.partner_limit = partner_limit
        d['infos'].append("Partner has been offered a credit limit of up to %.2f." % acct.partner_limit)
        # modify partner's actual iou limit if necessary
        partnerAcct = acct.get_partner_acct()
        oldPartnerIOULimit = partnerAcct.iou_limit
        if partnerAcct.my_limit != None:
          partnerAcct.iou_limit = min(partnerAcct.my_limit, acct.partner_limit)
        else:
          partnerAcct.iou_limit = acct.partner_limit
        if partnerAcct.iou_limit != oldPartnerIOULimit:
          partnerAcct.save()
          if acct.partner_limit == partnerAcct.iou_limit:
            d['infos'].pop() # clear mesage about partner being offered credit
          d['infos'].append("Partner's credit limit changed to %.2f." % partnerAcct.iou_limit)

    except ValueError:
      pass
    
    if request.POST['my_limit'] != '':
      try: # modify my max credit limit
        my_limit = float(request.POST['my_limit'])
        if my_limit < 0.0:
          d['infos'].append("Error: My credit limit must be a positive number or zero. '%s' is invalid." % request.POST['my_limit'])
          raise ValueError
        elif my_limit >= 1e12:
          d['infos'].append("Error: My credit limit amount is too large to be stored in the database.  Please enter another amount.")
          raise ValueError
        elif my_limit <= 1e-12:
          d['infos'].append("Error: My credit limit amount is too small to be stored in the database.  Please enter another amount.")
          raise ValueError
        
        if acct.my_limit != my_limit:
          acct.my_limit = my_limit
          d['infos'].append("You have chosen to automatically accept a credit limit of up to %.2f from your partner." % acct.my_limit)
      
      except ValueError:
        pass
    
    else: # set my limit to None = unlimited
      if acct.my_limit != None: # only change if not already None
        acct.my_limit = None
        d['infos'].append("You have chosen to automatically accept as high a credit limit as your partner offers you.")
    
    # modify my actual iou limit if necessary
    oldIOULimit = acct.iou_limit
    if acct.my_limit != None:
      acct.iou_limit = min(acct.my_limit, acct.get_partner_acct().partner_limit)
    else: # my_limit = unlimited
      acct.iou_limit = acct.get_partner_acct().partner_limit
    if acct.iou_limit != oldIOULimit:
      if acct.iou_limit == acct.my_limit:
        d['infos'].pop() # clear message about my maximum credit limit
      d['infos'].append("My credit limit changed to %.2f" % acct.iou_limit)

    acct.save()

    request.session['infos'] = d['infos']
    return HttpResponseRedirect('.')

  # render account details page
  # *** probably just need to pass acct, and maybe units
  d['cur'] = acct.get_shared_data().get_currency()
  d['acct'] = acct
  d['balance'] = acct.getBalance()
  #d['avgBalance'] = averageBalance(acct, 30)
  d['lowerLimit'] = acct.iou_limit
  d['upperLimit'] = acct.get_partner_acct().iou_limit
  d['displayBalance'] = d['balance']
  #d['displayAvgBalance'] = d['avgBalance']
  d['displayLowerLimit'] = d['lowerLimit']
  d['displayUpperLimit'] = d['upperLimit']
  if userNode.display_units_id and d['cur'].value:
    rate = d['cur'].value / userNode.get_display_units().value
    d['displayBalance'] *= rate
    #d['displayAvgBalance'] *= rate
    d['displayLowerLimit'] *= rate
    d['displayUpperLimit'] *= rate
  if d['cur'].value:
    d['convertibleUnits'] = currencyunits.get_list(value__gt=0.0, order_by=('long_name',));
  d['closeable'] = (float('%.2f' % d['balance']) == 0.0)
  
  # get links for payments touching this account
  ##d['links'] = paymentlinks.get_list(complex=(Q(payer_account__pk=acct.id) | Q(payer_account__partner_acct__pk=acct.id)), path__payment__id__ne=-1, order_by=('-t2.date',))
  ##links = paymentlinks.get_list(payer_account__id__in=(acct.id, acct.partner_acct_id),
  ##                              select={'date': 'select date from droplet_payments where id=pmt.id'},
  ##                              tables=['droplet_paymentpaths as path',
  ##                                      'droplet_payments as pmt'],
  ##                              where=['droplet_paymentlinks.path_id=path.id',
  ##                                     'path.payment_id=pmt.id'],
  ##                              order_by=('-date',))
  paymentTotal = 0.0
  months = []
  now = datetime.datetime.now()
  year = now.year
  month_num = now.month
  balance = acct.getBalance()
  highlight_row = False
  while True: # go month-by-month back to account creation
    month = {}
    links, interest, date = getLinksandInterestForMonth(acct, year, month_num)
    if year == now.year and month_num == now.month:
      interest += acct.getPresentInterest()
    else:
      month['date'] = date
    month['interest'] = interest
    month['balance'] = balance - paymentTotal
    if interest:
      month['highlight_row'] = highlight_row
      highlight_row = not highlight_row
    paymentTotal += interest
    months.append(month)

    entries = []
    last_entry = None
    for link in links:
      link.balance = balance - paymentTotal
      paymentTotal += link.amount
      if last_entry and last_entry.get_path().payment_id == link.get_path().payment_id:
        last_entry.amount += link.amount
      else:
        link.highlight_row = highlight_row
        highlight_row = not highlight_row
        entries.append(link)
        last_entry = link
    month['links'] = entries
    
    if month_num == 1:
      month_num = 12
      year = year -1
    else:
      month_num -= 1
    
    if datetime.datetime(year, month_num, 1) < acct.get_shared_data().created:
      break
  
  d['months'] = months
  d['paymentTotal'] = paymentTotal
  
  t = template_loader.get_template('accountDetail')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

def getLinksandInterestForMonth(acct, year, month):
  month_start = datetime.date(year=year, month=month, day=1)
  if month == 12:
    next_month_start = datetime.date(year=year + 1, month=1, day=1)
  else:
    next_month_start = datetime.date(year=year, month=month + 1, day=1)
  links = paymentlinks.get_list(payer_account__id__in=(acct.id, acct.partner_acct_id),
                                select={'date': 'select date from droplet_payments where id=pmt.id'},
                                tables=['droplet_paymentpaths as path',
                                        'droplet_payments as pmt'],
                                where=['droplet_paymentlinks.path_id=path.id',
                                       'path.payment_id=pmt.id',
                                       "pmt.date >= '%s'",
                                       "pmt.date < '%s'"],
                                params=[month_start, next_month_start],
                                order_by=('-date',))
  sum = 0.0
  for link in links:
    if link.payer_account_id == acct.id:
      link.interest = -link.interest # these amounts appear as negative (outgoing payments)
      link.amount = -link.amount
    sum += link.interest
    
  last_date = None
  if links:
    last_date = links[0].date
  return links, sum, last_date


# calculate time-weighted average of through payments
# added to sum of end-point payments
# period - in days
# **** FLAWED - can give different average to each partner ****
def averageBalance(acct, period):
  now = datetime.datetime.now()
  start = now - datetime.timedelta(days=period)
  links = paymentlinks.get_list(path__payment__date__gt=start, 
      payer_account__id__in=(acct.id, acct.partner_acct_id),
      order_by=('-t2.date',)) # order_by + join broken, should be '-droplet_payments.date' -- see Django ticket #1517
  syntheticBal = 0.0 # balance of thru-payments, moving backwards starting at zero
  averageBal = acct.getBalance() # what we're computing
  prevPmtTime = now
  prevPmtAmount = 0.0
  for link in links:
    # make sure link is for thru-payment
    if link.payer_account_id != acct.id or link.position != 1: # not payer
      if link.payer_account_id == acct.id or link.position != paymentlinks.get_count(path__pk=link.path_id): # not recipient
        syntheticBal -= prevPmtAmount
        linkTime = link.get_path().get_payment().date
        delta = prevPmtTime - linkTime
        # adjust by synthetic balance times fraction of period at that balance
        delta_days = delta.days + (delta.seconds + float(delta.microseconds) / 1000000) / 86400
        averageBal += syntheticBal * delta_days / period
        if link.payer_account_id == acct.id: # link is outgoing pmt on this acct
          prevPmtAmount = -link.amount
        else: # link is incoming pmt on this acct
          prevPmtAmount = link.amount
        prevPmtTime = linkTime
  
  # now factor in time from start of period to first paymentlink (last in list)
  syntheticBal -= prevPmtAmount
  delta = prevPmtTime - start
  delta_days = delta.days + (delta.seconds + float(delta.microseconds) / 1000000) / 86400
  averageBal += syntheticBal * delta_days / period
  
  return averageBal

def confirmClose(request, acctId):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  
  if not accounts.get_count(pk=acctId) > 0:
    return HttpResponseRedirect('/accounts/')
  
  d = {}
  d['acct'] = accounts.get_object(pk=acctId)
  if d['acct'].owner_id != userNode.id:
    return HttpResponseRedirect('/accounts/')
  t = template_loader.get_template('confirmClose')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

# for closing account and anything else we might want to do to the account
def accountAction(request, acctId, action):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/accounts/%s/' % acctId)
  
  if request.POST:
    if request.POST['submit'] == 'Cancel':
      return HttpResponseRedirect('/accounts/%s/' % acctId)
  
  if accounts.get_count(pk=acctId) == 0:
    return HttpResponseRedirect('/accounts/')
  
  acct = accounts.get_object(pk=acctId)
  if acct.owner_id != userNode.id:
    return HttpResponseRedirect('/accounts/')
  
  if action == 'close':
    sharedData = acct.get_shared_data()
    if abs(sharedData.balance) >= 0.005:
      request.session['infos'] = ["Account balance must be zero before it can be closed."]
      return HttpResponseRedirect('/accounts/%d/' % acct.id)
    
    # to be sure a transaction doesn't hit this account before it gets
    # deactivated, set limits to zero, save, and then check for zero balance again
    myIOULimit = acct.iou_limit
    partnerAcct = acct.get_partner_acct()
    partnerIOULimit = partnerAcct.iou_limit
    acct.iou_limit = 0.0
    acct.save()
    partnerAcct.iou_limit = 0.0
    partnerAcct.save()
    
    if abs(sharedData.balance) >= 0.005:
      request.session['infos'] = ["Account balance must be zero before it can be closed."]
      # restore limits on accts
      acct.iou_limit = myIOULimit
      acct.save()
      partnerAcct.iou_limit = partnerIOULimit
      partnerAcct.save()
      return HttpResponseRedirect('/accounts/%d/' % acct.id)
    
    # now it's impossible for a transaction to change the balance, and the balance is zero, so:
    sharedData.active = False
    sharedData.save()
    
    # notify user
    request.session['infos'] = ["Account %d (%s) closed." % (sharedData.id, acct.name)]
    
    # notify partner by email
    note = ''
    if request.POST:
      note = request.POST['note']
    t = template_loader.get_template('emailAcctClosed')
    c = DjangoContext(request, {'acct':partnerAcct, 'note':note})
    sendEmail("One of your accounts has been closed", t.render(c), acct.get_partner().getPrimaryEmail())
    pass
    
    return HttpResponseRedirect('/accounts/')

# process form from offerToEmails
def offer(request):
  """offer credit to multiple nodes"""
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  
  if not request.POST:
    return HttpResponseRedirect('/summary/')

  if request.POST['phase'] == 'make_offers': # check if offers are finalized, or if we're just listing emails and the form
    # extract info from POST dictionary
    rows = []
    for i in range(0, int(request.POST['count'])):
      row = Struct()
      row.email = request.POST['email_%d' % i] # already parsed as email addresses
      row.amount = request.POST['amount_%d' % i]
      row.currency = request.POST['currency_%d' % i]
      row.interest = request.POST['interest_%d' % i]
      rows.append(row)

    # find errors in entries
    error = False
    for row in rows:
      # ensure no double-offers or double-accounts
      emailList, errors = validateOfferEmails([row.email], userNode)
      if not emailList: # if row.email is in list, then offer is valid
        error = True
        row.error = errors[0] # only show first error
        continue # only one error per row

      # validate amounts as numbers
      try:
        row.amount = float(row.amount)
        if row.amount <= 0.0:
          raise ValueError
      except ValueError:
        error = True
        row.error = "Credit limit must be a number greater than zero."
      else:
        if row.amount >= 10**12:
          error = True
          row.error = "Amount is too large to be stored in the database.  Please enter another amount."
        elif row.amount <= 10**-12:
          error = True
          row.error = "Amount is too small to be stored in the database.  Please enter another amount."
      
      if error:
        continue # only one error per row
      
      try:
        row.interest = float(row.interest)
      except ValueError:
        error = True
        row.error = "Interest rate must be a number."
    
    note = request.POST['note']
    if note == '': note = None
    
    if error: # show form again with errors and values still there (in row objects)
      d = {'rows':rows, 'count':len(rows)}
      d['infos'] = ["One of more of your entries had errors. Please make corrections below."]
    else: # create and save offers, and send out emails
      infos = []
      infos.append('The system has recorded your offers of credit and is sending out invitation emails.  <a href="/accounts/#offers">Click here to see your offers.</a>')
      #infos.append("If the invitation does not arrive, it may be in the destination's spam folder...")
      for row in rows:
        currency = currencyunits.get_object(short_name__exact=row.currency)
        offer = offers.Offer(initiator=userNode, recipient_email=row.email, 
                             amount=row.amount, currency=currency, interest_rate=row.interest)
        offer.save()

        # send out email
        t = template_loader.get_template('emailInvitation')
        c = DjangoContext(request, {'row':row, 'userNode':userNode, 'currency':currency, 'note':note})
        sender = '"%s" <%s>' % (userNode.name, userNode.getPrimaryEmail())
        if not sendEmail('Invitation to Rain Droplet', t.render(c), row.email, sender=sender, attempts=2, includeServiceName=False):
          infos.append("An email could not be sent to %s, but your offer has been stored. Please contact %s manually." % (escape(row.email), escape(row.email)))
      request.session['infos'] = infos
      return HttpResponseRedirect('/accounts/')

  else: # else, list emails + form
    emails = parseEmails(request.POST['emailText'])
    emails, infos = validateOfferEmails(emails, userNode)
    if emails == [] and infos == []:
      infos.append("You didn't enter any valid email addresses.")
    rows = []
    for email in emails:
      row = Struct()
      row.email = email
      row.currency = None
      rows.append(row)
    d = {'infos':infos, 'rows':rows, 'count':len(rows)}
  
  d['currencies'] = currencyunits.get_list(order_by=('long_name',))
  t = template_loader.get_template('offer')
  c = DjangoContext(request, d)
  return HttpResponse(t.render(c))

def validateOfferEmails(emails, userNode):
  infos = []
  for email in emails[:]: # use slice operator to make a copy of emails because it will be modified during loop
    remove = False
    
    # check if email address belongs to someone
    if emailaddrs.get_count(email__exact=email) > 0:
      emailAddr = emailaddrs.get_object(email__exact=email)
      
      # check if user already has account with owner of this email address
      if accounts.get_count(owner__pk=userNode.id, partner__pk=emailAddr.node_id, shared_data__active__exact=True) != 0:
        remove = True
        acct = accounts.get_object(owner__pk=userNode.id, partner__pk=emailAddr.node_id, shared_data__active__exact=True)
        infos.append('You already have an account with %s. <a href="/accounts/%d/">You can modify it here.</a>' % (escape(email), acct.id))
      
      # check if email address has made an offer to this user
      if offers.get_count(recipient_email__in=getEmails(userNode), initiator__pk=emailAddr.node_id) != 0:
        remove = True
        offer = offers.get_list(recipient_email__in=getEmails(userNode), initiator__pk=emailAddr.node_id)[0]
        infos.append('%s has already offered <b>you</b> %s %.2f credit. <a href="/offers/%d/accept/">Click here to accept.</a>' % (escape(email), offer.get_currency().short_name, offer.amount, offer.id))
        
    # check if user has already made an offer to this email address
    if offers.get_count(initiator__pk=userNode.id, recipient_email__exact=email) != 0:
      remove = True
      offer = offers.get_object(initiator__pk=userNode.id, recipient_email__exact=email)
      infos.append('You have already made an offer to %s. <a href="/offers/%d/withdraw/">Click here to withdraw it.</a>' % (escape(email), offer.id))
    
    # remove userNode's email if present
    if email in getEmails(userNode, confirmed=False):
      remove = True
      infos.append("You can't connect to yourself! Silly monkey.")
    
    if remove:
      emails.remove(email)

  return emails, infos

def rejectionNote(request, offerId):
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  
  if offers.get_count(pk=offerId) == 0:
    request.session['infos'] = ["Offer has been withdrawn by the sender."]
    return HttpResponseRedirect('/summary/')
  
  offer = offers.get_object(pk=offerId)
  if offer.recipient_email not in getEmails(userNode):
    return HttpResponseRedirect('/summary/')
  
  t = template_loader.get_template('rejectionNote')
  c = DjangoContext(request, {'offer':offer})
  return HttpResponse(t.render(c))
  

def offerAction(request, offerId, action):
  """Accept or reject somebodys offer of credit"""
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  
  message = ''
  if offers.get_count(pk=offerId) == 1:
    offer = offers.get_object(pk=offerId)
  else: # offer has been deleted since action link was displayed to user
    if action == 'accept' or action == 'reject':
      message = "Offer has been withdrawn and cannot be accepted or rejected."
    else:
      message = "Offer has already been acted on and cannot be withdrawn."
    request.session['infos'] = [message]
    return HttpResponseRedirect('/summary/')

  # check the offer is for the logged-in user if action is accept or reject
  if (action == 'accept' or action == 'reject') and \
      offer.recipient_email not in getEmails(userNode):
    return HttpResponseRedirect('/summary/')

  if action == 'accept':
    initiator = offer.get_initiator()
    # check for existing account between nodes
    # *** temporary until multiple accounts permitted
    if accounts.get_count(owner__pk=userNode.id, partner__pk=initiator.id, shared_data__active__exact=True) > 0:
      request.session['infos'] = ["Account already exists with %s.  Multiple accounts with the same neighbour are not permitted (yet)." % escape(initiator.name)]
      offer.delete()
      return HttpResponseRedirect('/summary/')
    currencyId = offer.currency_id
    shared = sharedaccountdatas.SharedAccountData(currency_id=currencyId, balance=0.0, active=True,
                                                  interest_rate=offer.interest_rate/100.0)
    shared.save()
    acct1 = accounts.Account(owner=initiator, partner=userNode, shared_data=shared, balance_multiplier=1, iou_limit=0.0, partner_limit=offer.amount, name=userNode.name)
    acct1.save()
    acct2 = accounts.Account(owner=userNode, partner=initiator, shared_data=shared, balance_multiplier=-1, iou_limit=offer.amount, partner_limit=0.0, partner_acct=acct1, name=initiator.name)
    acct2.save()
    acct1.partner_acct_id = acct2.id
    acct1.save()
    offer.delete()
    
    # shuffle node locations *** slow - consider doing as a regular script instead
    circle = routing.Circle()
    circle.shuffle(10 * nodes.get_count())
    
    # send email informing initiator that offer was accepted
    t = template_loader.get_template('emailInvitationAccepted')
    c = DjangoContext(request, {'offer':offer, 'acct':acct1})
    sendEmail("Invitation accepted", t.render(c), initiator.getPrimaryEmail())
    
    request.session['infos'] = ["This is your new connection account with %s.  Please take the time to give your account a meaningful name and assign your new neighbour a credit limit." % escape(initiator.getPrimaryEmail())]
    return HttpResponseRedirect('/accounts/%d/' % acct2.id)
  elif action == 'reject':
    offer.delete()
    
    note = ''
    if request.POST:
      note = request.POST['note']
    
    # send email informing initiator that offer was rejected
    t = template_loader.get_template('emailInvitationRejected')
    c = DjangoContext(request, {'offer':offer, 'note':note})
    sendEmail("Your invitation was not accepted", t.render(c), offer.get_initiator().getPrimaryEmail())
    
    message = "Offer from %s rejected." % escape(offer.get_initiator().getPrimaryEmail())
  elif action == 'withdraw':
    if userNode != offer.get_initiator():
      return HttpResponseRedirect('/accounts/')
    offer.delete()
    message = "Offer to %s withdrawn." % offer.recipient_email

  if message:
    request.session['infos'] = [message]
  return HttpResponseRedirect('/accounts/')


def interestRateAction(request, acctId, action):
  "Accept or reject a rate-change proposal on an account."
  userNode = checkLogin(request)
  if not userNode: return HttpResponseRedirect('/login/?redirect=%s' % request.path)
  
  if accounts.get_count(owner__pk=userNode.id, pk=acctId) == 0:
    return HttpResponseRedirect('/accounts/')
  
  acct = accounts.get_object(pk=acctId)
  shared_data = acct.get_shared_data()
  if shared_data.proposed_rate <> None and shared_data.node_to_confirm_id == userNode.id:
    if action == 'accept':
      accept_proposed_rate(request, shared_data)
    elif action == 'reject':
      reject_proposed_rate(request, shared_data)
  
  return HttpResponseRedirect('/accounts/%d/' % acct.id)

def accept_proposed_rate(request, shared_data):
  # this is node_to_confirm's account
  acct = accounts.get_object(shared_data__pk=shared_data.id, owner__pk=shared_data.node_to_confirm_id)
  
  # *** for django 0.95: wrap in transaction
  pmt = None
  try:
    if acct.getBalance() and shared_data.interest_rate: # otherwise no update necessary

      pmt = payments.Payment(payer=acct.get_owner(),
                             payer_email=acct.get_owner().getPrimaryEmail(),
                             recipient=acct.get_partner(),
                             recipient_email=acct.get_partner().getPrimaryEmail(),
                             currency=shared_data.get_currency(),
                             amount=0.0,
                             status='PE',
                             description='Interest rate change: %s%% to %s%%.' % \
                             (shared_data.displayRate(), shared_data.displayProposedRate()))
      pmt.save() # set pmt time
      pmt = payments.get_object(pk=pmt.id) # reload because django doesn't update date that is auto-set on save
      
      interest = computeInterest(acct.getBalance(), shared_data.interest_rate, shared_data.last_update, pmt.date)
      
      path = paymentpaths.PaymentPath(payment=pmt, amount=pmt.amount)
      path.save()
      
      link = paymentlinks.PaymentLink(path=path, payer_account=acct, position=1, 
                                      amount=pmt.amount, interest=-interest, 
                                      interest_rate=shared_data.interest_rate)
      link.save()
      
      shared_data.balance += interest * acct.balance_multiplier
      
    if pmt:
      shared_data.last_update = pmt.date
    else:
      shared_data.last_update = datetime.datetime.now()
    shared_data.interest_rate = shared_data.proposed_rate
    shared_data.proposed_rate = None
    shared_data.node_to_confirm_id = None
    shared_data.save()
  
    if pmt: # for now, try to be safe by not marking payment as OK until the end
      pmt.status = 'OK'
      pmt.save()
  
    # *** commit transaction here
  except Exception, e:
    # *** rollback transaction here
    raise
  
  t = template_loader.get_template('acceptRateProposal')
  c = DjangoContext(request, {'acct': acct.get_partner_acct()})
  sendEmail("Interest rate proposal accepted", t.render(c), acct.get_partner().getPrimaryEmail())

def reject_proposed_rate(request, shared_data):
  
  proposed_rate = shared_data.proposed_rate * 100.0
  partner_acct = accounts.get_object(shared_data__pk=shared_data.id, owner__id__ne=shared_data.node_to_confirm_id)
  shared_data.proposed_rate = None
  shared_data.node_to_confirm_id = None
  shared_data.save()
  
  t = template_loader.get_template('rejectRateProposal')
  c = DjangoContext(request, {'acct': partner_acct, 'proposed_rate': proposed_rate})
  sendEmail("Interest rate proposal not accepted", t.render(c), partner_acct.get_owner().getPrimaryEmail())

def propose_rate(request, shared_data, interest_rate, node_to_confirm):
  acct_to_confirm = accounts.get_object(shared_data__pk=shared_data.id, owner__pk=node_to_confirm.id)
  shared_data.proposed_rate = interest_rate
  shared_data.node_to_confirm_id = node_to_confirm.id
  shared_data.save()
  
  t = template_loader.get_template('interestRateProposal')
  c = DjangoContext(request, {'acct': acct_to_confirm})
  if sendEmail("Interest rate proposal", t.render(c), node_to_confirm.getPrimaryEmail(), attempts=2):
    return "Your partner has been notified of your proposed rate change."
  else:
    return "Your rate proposal has been recorded, however, the system could not send an email notifying them.  Please contact them manually."
